/* data.c, a part of Twig image gallery program
 * (c) 2003-2006 Twibright Labs
 * Twig is released under GPL.
 */

#ifndef _DATA_H_
#define _DATA_H_

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>

extern unsigned char *progname;

void *mem_alloc(size_t size)
{
	void *rv;
	if (!size) return NULL;
	rv=malloc(size);
	if (rv) return rv;
	fprintf(stderr,"%s: Can't malloc %d bytes\n",progname, size);
	exit(1);
}

FILE *sfopen(const char *filename, const char *mode)
{
	FILE *f;

	f=fopen(filename, mode);
	if (!f){
		fprintf(stderr,"%s: Can't open file %s: ",progname, filename);
		perror("");
		exit(1);
	}
	return f;
}

void sfclose(FILE *f)
{
	int rv;

	rv=fclose(f);
	if (rv){
		fprintf(stderr,"%s: Can't close file: ",progname);
		perror("");
		exit(1);
	}
}

/* sprintf with automatic reallocation that returns the result. The
 * result has to be freed by free() aferwards. */
unsigned char * ssprintf(unsigned char *format, ...)
{
	int size=128;
	unsigned char *str;
	int n;
	va_list ap;

	str=mem_alloc(size);
	while(1){
		va_start(ap, format);
		n=vsnprintf(str, size, format, ap);
		va_end(ap);
		if (n>-1&&n<size) return str;
		if (n>-1)size=n+1;
		else size*=2;
		if (!(str=realloc(str, size))){
				fprintf(stderr,"%s: Can't realloc\n",progname);
				exit(1);
		}
	}
}

void save_8_bytes(unsigned char *dest, unsigned long long id)
{
	dest[0]=id;
	dest[1]=id>>8;
	dest[2]=id>>16;
	dest[3]=id>>24;
	dest[4]=id>>32;
	dest[5]=id>>40;
	dest[6]=id>>48;
	dest[7]=id>>56;
}

unsigned long long load_8_bytes(unsigned char *src)
{
	unsigned long long retval;

	retval=src[0];
	retval|=src[1]<<8;
	retval|=src[2]<<16;
	retval|=src[3]<<24;
	retval|=((unsigned long long)src[4])<<32;
	retval|=((unsigned long long)src[5])<<40;
	retval|=((unsigned long long)src[6])<<48;
	retval|=((unsigned long long)src[7])<<56;
	return retval;
}

void save_4_bytes(unsigned char *dest, unsigned long time)
{
	*dest=time;
	dest[1]=time>>8;
	dest[2]=time>>16;
	dest[3]=time>>24;
}

unsigned long load_4_bytes(unsigned char *from)
{
	unsigned long retval;
	
	retval=*from;
	retval|=from[1]<<8;
	retval|=from[2]<<16;
	retval|=from[3]<<24;

	return retval;
}
#endif /* #ifndef _DATA_H_ */
