/* manager.c, the main C "script" for Twig image gallery program
 * (c) 2003-2006 Twibright Labs
 * Twig is released under GPL.
 */

#include <gdbm.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <unistd.h>

#include "data.h"

GDBM_FILE dbf;

int remake_html; /* 1 means that even untouched images will have their
		    HTML remade (but without regenerating thumbnails) */
unsigned long long free_id;
unsigned char *progname="manager";
const unsigned char charset[]="utf-8";
const unsigned char doctype[]=
		"<!DOCTYPE html PUBLIC "
		"\"-//W3C//DTD HTML 4.01 Transitional//EN\""
		" \"http://www.w3.org/TR/html4/loose.dtd\">\n";
const unsigned char copyright[]="<hr><center>Generated by "
	"<a href=\"http://ronja.twibright.com/utils/twig/\">Twig</a>"
	" from\n"
	"<a href=\"http://www.twibright.com\">Twibright"
	" Labs</a></center></body></html>\n";
const unsigned char colors[]="bgcolor=\"#000000\" text=\"#7fffff\"\n"
	"link=\"#ffff00\" vlink=\"#00ff00\" alink=\"#c000ff\"";
const unsigned char hdr_left[]="<table width=\"100%\">\n"
"<tr><td align=left valign=top><a href=\"http://ronja.twibright.com/\"><img\n"
"src=\"http://ronja.twibright.com/grx/logo/ronja_logo2.png\" width=78 height=93\n"
"border=0 alt=\"Twibright Ronja\"></a>\n"
"<td align=center valign=bottom><h1>";

const unsigned char hdr_right[]="</h1>\n"
"<td align=right valign=top><a href=\"http://twibright.com/\"><img\n"
"src=\"http://ronja.twibright.com/grx/logo/tbl3.png\"\n"
"alt=\"Twibright Labs\" width=286 height=72 border=0></a>\n"
"</table>\n";
unsigned char switchcode;
unsigned long long this_directory_id;
unsigned char *this_directory_index_filename;
int table_written;

FILE * big_file;
int col=0;
int cols=14; /* For the lvl0 thumbnails */
int cols2=5; /* For the lvl1 thumbnails in directory listings */

unsigned char * make_shortname(unsigned char *filename)
{
	int length=strlen(filename);
	unsigned char *ptr=filename;
	unsigned char *retval;

	while (*ptr&&*ptr!='/'){
		ptr++;
		length--;
	}
	if (!*ptr){
		ptr=filename;
		length=strlen(filename);
	}else{
		ptr++;
		length--;
	}
	retval=mem_alloc(length+1);
	memcpy(retval, ptr, length+1);	
	return retval;
}

unsigned char *make_lastname(unsigned char *filename)
{
	int len=strlen(filename);
	unsigned char *ptr=filename+len-1;
	unsigned char *rv;
	
	ptr=strrchr(filename, '/');
	if (!ptr) ptr=filename;
	else ptr++;
	rv=mem_alloc(strlen(ptr)+1);
	strcpy(rv,ptr);
	return rv;
}

void add_to_big(unsigned char *filename, unsigned long long id)
{
	static const unsigned char str1[]=
		"<td valign=middle align=center>\n"
		"<a href=\"%llx.html\">\n"
		"<img src=\"lvl0/%llx.jpg\" border=0 "
		"alt=\"L0 thumbnail\">\n"
		"</a></td>\n";

	if (!col) fputs("<tr>",big_file);
	fprintf(big_file,str1,id,id);
	col++;
	if (col>=cols){
		col=0;
		fputs("</tr>",big_file);
	}
}

void add_reverse(unsigned long long id, unsigned char *filename)
{
	unsigned char t[8];
	datum key={t, sizeof(t)};
	datum d;

	save_8_bytes(t, id);
	d.dptr=filename;
	d.dsize=strlen(filename)+1;
	gdbm_store(dbf, key, d, GDBM_REPLACE);
}

int isjpeg(unsigned char *string)
{
	int len;

	len=strlen(string);
	if (len<3) return 0;
	if (!strcasecmp(string+len-3,"jpg")) return 1;
	if (len<4) return 0;
	if (!strcasecmp(string+len-3,"jpeg")) return 1;
	if (!strcasecmp(string+len-3,"jfif")) return 1;
	return 0;
}

int is_ronja(unsigned char *shortname)
{
	unsigned char pattern[]="ronja";
	
	return !memcmp(shortname, pattern, sizeof(pattern)-1);
}

void my_system(const unsigned char *cmdline)
{
	int retval;
	
	retval=system(cmdline);
	retval=WEXITSTATUS(retval);
	if (retval){
		fprintf(stderr,"Command %s generated nonzero exit status"
			"%d.\r\n",cmdline, retval);
	}
}

void move_thumbnail(unsigned long long id, int level)
{
	static unsigned char dst[4096];
	static unsigned char src[4096];

	sprintf(dst,"tns/lvl%d/%llx.jpg",level,id);
	sprintf(src,"temptns/lvl%d/%llx.jpg",level,id);
	if (rename(src, dst)){
		fprintf(stderr,"%s: cannot rename file %s to %s:",
				progname, src, dst);
		perror("");
	}
}

void fputs_esc(unsigned char *string, FILE *f)
{
	unsigned char c;

	while((c=*string++))
		switch(c){
			default:
				putc(c, f);
				break;
			case '>':
				fputs("&gt;",f);
				break;
			case '<':
				fputs("&lt;",f);
				break;
			case '&':
				fputs("&amp;",f);
				break;
			case '"':
				fputs("&quot;",f);
				break;
		}
}

/* Remakes the thumbnails and image's infopage */
void remake_image(unsigned char *filename, unsigned long long id,
		int skip_thumbs)
{

	unsigned char *shortname;
	static unsigned char output_filename[4096];
	static unsigned char commandline[4096];
	int i;
	FILE *f;

	shortname=make_shortname(filename);
	sprintf(output_filename,"temptns/%llx.html",id);
	f=sfopen(output_filename,"w");
	
	fputs(doctype, f);
	fprintf(f,
		"<html><head>\n"
		"<meta http-equiv=\"Content-Type\" content=\"text/html;"
		" charset=%s\">\n"
		"<title>[",charset);
	fprintf(f,"%llx] :: ",id);
	fputs_esc(shortname, f);
	fputs("\n</title></head>",f);
	fprintf(f,"<body %s>\n",colors);
	if (is_ronja(shortname)){
		fputs(hdr_left,f);
		fprintf(f,"[%llx] ",id);
		fputs_esc(shortname,f);
		fputs(hdr_right,f);
	}else{
		fprintf(f,
			"<h1>[%llx] ",id);
		fputs_esc(shortname, f);
		fputs("</h1>\n",f);
	}
	fputs(
		"<p><table width=\"100%\" border=0><tr>"
		"<td align=right valign=middle>\n"
		"<a href=\"../",f);
	fputs_esc(filename,f);
	fprintf(f,"\">\n"
		"<img src=\"lvl2/%llx.jpg\""
		" alt=\"L2 thumbnail\">\n"
		"</a></td><td align=left valign=middle>\n"
		"<a href=\"../",id);
	fputs_esc(filename, f);
	fputs("\">",f);

	if (!skip_thumbs)
	{
		sprintf(commandline,"convert -size 256x256 -resize 256x256 \"%s\""
			" +profile \"*\" temptns/lvl2/%llx.jpg", filename, id);
		my_system(commandline);
		sprintf(commandline,"convert -size 128x128 -resize 128x128"
			" +profile \"*\" temptns/lvl2/%llx.jpg"
			" temptns/lvl1/%llx.jpg", id, id);
		my_system(commandline);
		sprintf(commandline,"convert -size 64x64 -resize 64x64"
			" +profile \"*\" temptns/lvl1/%llx.jpg"
			" temptns/lvl0/%llx.jpg", id, id);
		my_system(commandline);

		sprintf(commandline,"convert -resize 70.7107%%"
			" +profile \"*\" \"%s\""
			" temptns/lvl5/%llx.jpg", filename, id);
		my_system(commandline);
		sprintf(commandline,"convert -resize 70.7107%%"
			" +profile \"*\" temptns/lvl5/%llx.jpg"
			" temptns/lvl4/%llx.jpg", id, id);
		my_system(commandline);
		sprintf(commandline,"convert -resize 70.7107%%"
			" +profile \"*\" temptns/lvl4/%llx.jpg"
			" temptns/lvl3/%llx.jpg", id, id);
		my_system(commandline);

		/* Move the thumbnails from temptns to tns. */
		for (i=0;i<=5;i++){
			move_thumbnail(id, i);
		}
	}

	sfclose(f);

	sprintf(commandline,"identify -format '%%m<br>%%w x %%h<br>%%b</a></td>\n<td"
		" align=left valign=top rowspan=4><pre>"
		/* EXIF data headers begin */
		"Taken at     : %%[EXIF:DateTime]\n"
		"Camera       : %%[EXIF:Make] %%[EXIF:Model]\n"
		"Original size: %%[EXIF:ExifImageWidth"
			"] x %%[EXIF:ExifImageLength]\n"
		"' \"%s\" >> %s",filename, output_filename);
	my_system(commandline);

	sprintf(commandline,"jhead \"%s\" | sed -n -e '/File size/p;"
		"/^Flash used/p;"
		"/^Focal length/p;"
		"/^Exposure time/p;"
		"/^Aperture/p;"
		"/^Metering Mode/p;"
		"/^Exposure  /p;"
		"/^ISO equiv./p;"
		"/^CCD width/p;"
		"/^Focust dist./p;"
		"/^Whitebalance/p;"
		"' >> \"%s\"",filename, output_filename);
	my_system(commandline);

	f=sfopen(output_filename,"a");
	fprintf(f,"</pre></td></tr>\n<tr><td align=right valign=middle>"
		"<a href=\"lvl5/%llx.jpg\"><img src="
		"\"lvl0/%llx.jpg\""
		" alt=\"L0 thumbnail\"></a></td><td "
	       "valign=middle"
		" align=left>\n"
		"<a href=\"lvl5/%llx.jpg\">",id, id, id);
	sfclose(f);
		
	sprintf(commandline,"identify -format '%%w x %%h<br>%%b' tns/lvl5/%llx.jpg"
			" >> %s",id, output_filename);
	my_system(commandline);

	f=sfopen(output_filename,"a");
	fprintf(f,"</a></td></tr>\n<tr><td align=right valign=middle>"
		"<a href=\"lvl4/%llx.jpg\"><img src="
		"\"lvl0/%llx.jpg\""
		" alt=\"L0 thumbnail\"></a></td><td align=left"
		" valign=middle>\n"
		"<a href=\"lvl4/%llx.jpg\">",id, id, id);
	sfclose(f);
		
	sprintf(commandline,"identify -format '%%w x %%h<br>%%b"
		"</a></td></tr>\n' tns/lvl4/%llx.jpg"
			" >> %s",id, output_filename);
	my_system(commandline);

	f=sfopen(output_filename,"a");
	fprintf(f,"<tr><td align=right valign=middle>"
			"<a href=\"lvl3/%llx.jpg\"><img src="
		"\"lvl0/%llx.jpg\""
		" alt=\"L0 thumbnail\"></a></td><td align=left"
		" valign=middle>\n"
		"<a href=\"lvl3/%llx.jpg\">",id, id, id);
	sfclose(f);
		
	sprintf(commandline,"identify -format '%%w x %%h<br>%%b"
		"</a></td></tr></table>' tns/lvl3/%llx.jpg"
			" >> %s",id, output_filename);
	my_system(commandline);

	sprintf(commandline,"%s \"%s\" | sed -e '"
		"s/\\&/\\&amp;/g;s/</\\&lt;/g;s/>/\\&gt;/g;s/"
		"^/<p>/g' >> %s", 
		(isjpeg(filename)?"rdjpgcom":"identify -format '%c'"),
		filename, output_filename);
	my_system(commandline);

	f=fopen(output_filename,"a");
	fputs(copyright,f);

	free(shortname);
	sfclose(f);

	add_reverse(id, filename);
}


/* Byte 0: time (4 bytes)
 * Byte 4: id (8 bytes)
 * return value: 0=OK
 *               1=key doesn't exist
 */
int parse_record(unsigned long long *id, unsigned long *time,
	datum key)
{
	datum d=gdbm_fetch(dbf, key);

	if (!d.dptr) return 1;
	if (d.dsize!=12){
		fprintf(stderr,"Internal error: data size is not 13 bytes\n");
		exit(1);
	}
	*time=load_4_bytes(d.dptr);
	*id=load_8_bytes(d.dptr+4);
	return 0;
}

/* The "id\0" name has the time record empty */
void store_record(unsigned long long id, unsigned long time,
	datum key)
{
	unsigned char t[12];
	datum d={t, 12};
	save_4_bytes(t, time);
	save_8_bytes(t+4, id);
	if (gdbm_store(dbf, key, d, GDBM_REPLACE))
	{
		fprintf(stderr,"Can't store into database\n");
		exit(2);
	}
}

unsigned long filetime(unsigned char *filename)
{
	struct stat buf;

	stat(filename, &buf);
	return buf.st_mtime;
}

void begin_dir_file(unsigned long long id, unsigned char *filename)
{
	unsigned char output_filename[4096];
	unsigned char *shortname=make_shortname(filename);
	FILE *f;

	sprintf(output_filename,"temptns/%llx.html",id);
	f=fopen(output_filename,"w");
	if (!f){
		perror(progname);
		exit(3);
	}
	fputs(doctype,f);
	fprintf(f,
		"<html><head>\n"
		"<meta http-equiv=\"Content-Type\" content=\"text/html;"
		" charset=%s\">\n"
		"<title>[",charset);
	fprintf(f,"%llx] ",id);
	fputs_esc(shortname, f);
	fprintf(f,"</title></head>\n"
		"<body %s>\n",colors);
	if (is_ronja(shortname)){
		fputs(hdr_left,f);
		fprintf(f,"[%llx] ",id);
		fputs_esc(shortname, f);
		fputs(hdr_right,f);
	}else{
		fprintf(f,"<h1>[%llx] ",id);
		fputs_esc(shortname, f);
		fputs("</h1>\n",f);
	}

	free(shortname);
	sfclose(f);
	
}

/* Takes a list of directories on stdin. */
void load_directories(void)
{
	int len;
	unsigned char buffer[4096];
	datum key;

	key.dptr=buffer;
	while(fgets(buffer, sizeof(buffer), stdin))
	{
		len=strlen(buffer);
		if (buffer[len-1]=='\n'){
		       buffer[len-1]=0;
		       len--;
		}
		if (!len) continue;
		key.dsize=len;
		if (!gdbm_exists(dbf, key)){
			fprintf(stderr,"Assigning %llx to directory %s\n"
				,free_id, buffer);
			store_record(free_id, 0, key);
			begin_dir_file(free_id, buffer);
			free_id++;
		}else{
			unsigned long long id;
			unsigned long database_time;

			parse_record(&id, &database_time, key);
			begin_dir_file(id, buffer);
		}
	}
}

unsigned char *reverse(unsigned long long id)
{
	unsigned char t[8];
	datum key={t, sizeof(t)};
	datum d;

	save_8_bytes(t, id);
	d=gdbm_fetch(dbf, key);
	if (!d.dptr){
		fprintf(stderr,"%s: Can't reverse lookup id=%llx\n", progname, id);
		exit(4);
	}
	return d.dptr;
}

/* Takes a list of image files on stdin */
void load_files(void)
{
	int len;
	datum key;
	unsigned char buffer[4096];
	unsigned long file_time;

	key.dptr=buffer;
	while(fgets(buffer, sizeof(buffer), stdin))
	{
		len=strlen(buffer);
		if (buffer[len-1]=='\n'){
		       buffer[len-1]=0;
		       len--;
		}
		if (!len) continue;
		key.dsize=len;
		file_time=filetime(buffer);
		if (!gdbm_exists(dbf, key)){
			fprintf(stderr,"Making new info page for %s, assigning "
				"%llx\n",buffer,free_id);
			add_to_big(buffer,free_id);

			/* Store the new record and mark it as dirty */
			store_record(free_id, file_time, key);

			/* Make new thumbnails and page for the image */
			/* This call also increases free_id. */	
			remake_image(buffer, free_id, 0);
			/* Do not skip thumbs */
			free_id++;

		}else{
			unsigned long long id;
			unsigned long database_time;

			parse_record(&id, &database_time, key);
			add_to_big(buffer, id);

			if (database_time!=file_time){
				fprintf(stderr,"Remaking info page for %s"
					" [%llx]\n", buffer, id);
				store_record(id, file_time, key);
				remake_image(buffer, id, 0);
				/* Do not skip thumbs */
			}else if (remake_html)
			{
				fprintf(stderr,"Remaking HTML for "
					"%s [%llx]\n",buffer, id);
				remake_image(buffer, id, 1);
				/* Skip thumbs */
			}
		}
			
			
	}
}

/* Something to do with the index file */
void init_big_file(void)
{
	big_file=fopen("temptns/index.html","w");
	if (!big_file) {
		perror(progname);
		exit(5);
	}
	fputs(doctype,big_file);
	fprintf(big_file,
		"<html><head>\n"
		"<meta http-equiv=\"Content-Type\" content=\"text/html;"
		" charset=%s\">\n"
		"<title>Twibright Images</title></head>\n"
		"<body %s>\n",charset, colors);
	fputs(
		"<h1>Twibright Images Overall View</h1>\n"
		"<table border=1 cellpadding=0 cellspacing=0>\n"
		,big_file);
}

void close_big_file(void)
{
	if (col) fputs("</tr>\n",big_file);
	fprintf(big_file
		,"</table>%s",copyright);
	sfclose(big_file);
}

/* free_id is saved as 8 bytes */
void load_free_id(void)
{
	datum key={"id",3};
	unsigned long time;

	if (parse_record(&free_id, &time, key)) free_id=0;
	fprintf(stderr,"Next free id is %llx\n",free_id);
}

void save_free_id(void)
{
	datum key={"id",3};
	fprintf(stderr,"Next free id is %llx\n",free_id);
	store_record(free_id, 0, key);
}

/* 0 = directories
 * 1 = files 
 * Takes a list of image files / directories on stdin
 */
void process_subentries(int switchcode)
{
	int len;
	unsigned char buffer[4096];
	datum key;
	unsigned long long id;
	unsigned long time;
	FILE *f;
	unsigned char *last_name;
	unsigned char *commandline;
	unsigned char *rvs;

	f=sfopen(this_directory_index_filename,"a");

	key.dptr=buffer;
	while(fgets(buffer, sizeof(buffer), stdin))
	{
		len=strlen(buffer);
		if (buffer[len-1]=='\n'){
		       buffer[len-1]=0;
		       len--;
		}
		if (!len) continue;
		key.dsize=len;
		parse_record(&id, &time, key);

		last_name=make_lastname(buffer);
		if (switchcode){
			/* Doing the bottom part -- files */
			if (!table_written){
				table_written=1;
				fprintf(f,"<center><table border=1"
					" width=\"100%%\">\n");
			}
			if (!col) fprintf(f,"<tr>\n");
			fprintf(f,"<td align=center valign=middle>"
					"<table border=0>\n"
					"<tr><td"
					" valign=bottom align=center"
					">[%llx] ",id);
			fputs_esc(last_name,f);
			fprintf(f,"</td></tr>\n"
					"<tr><td"
					" valign=middle align=center"
					"><a href=\"%llx.html\">"
					"<img border=0 src=\"lvl1/%llx.jpg\""
					" alt=\"L1 thumbnail\""
					"></a></td></tr>\n<tr><td"
					" valign=top align=center>",
					id, id);
			sfclose(f);
			rvs=reverse(id);
			commandline=ssprintf("%s \"%s\" |"
				" head -n 1 | sed -e '"
				"s/\\&/\\&amp;/g;s/</\\&"
				"lt;/g;s/>/\\&gt;/g' >> %s",
				(isjpeg(rvs)?"rdjpgcom":"identify -format"
					 " '%c'"),
				rvs,
				this_directory_index_filename);
			my_system(commandline);
			f=sfopen(this_directory_index_filename,"a");
			fprintf(f,"</td></tr></table></td>");
			col++;
			if (col==cols2){
				col=0;
				fprintf(f,"</tr>\n");
			}
		}else{
			/* Doing the top part -- directories */
			if (!table_written){
				table_written=1;
				fprintf(f,"<table border=0"
					" width=\"100%%\">\n");
			}
			fprintf(f,"<tr><td align=right valign=middle>"
				"<a href=\"%llx.html\">"
				"<img src=\"../folder.gif\""
				"alt=\"Folder\"></a></td>\n"
				"<td align=left valign=middle"
				"><a href=\"%llx.html\">",id, id);
			fputs_esc(last_name,f);
			fputs("</a></td></tr>\n",f);
		}
		free(last_name);
	}
	if (switchcode){
		/* Doing the lower part -- the files */
		if (table_written){
			if (col){
				/* We are not at the beginning of the row */
				fputs("</tr>\n</table></center>",f);
			}else{
				/* We are at the end of the row */
				fputs("</table></center>",f);
			}
		}
		fputs(copyright,f);
	}else{
		/* The upper part -- the directories */
		if (table_written) fprintf(f,"</table>\n");
	}
	fclose(f);
}

/* Switchcodes:
 * 1 Scanning for new directories
 * 2 Scanning for new/changed files. Creates a file html page for every file.
 * 3
 * 4
 */
int main(int argc, char **argv)
{
	if (getenv("TWIG_REMAKE_HTML")) remake_html=1;
	if (argc<2){
		fprintf(stderr,"usage: manager <switchcode>\n");
		exit(0);
	}
	switchcode=*(argv[1]);
	dbf=gdbm_open("temptns/database",4096,GDBM_WRCREAT, 0644, NULL);
	switch(switchcode){
		case '1':
		fprintf(stderr,"%s: Scanning for new directories.\n",progname);
		load_free_id();
		load_directories();
		save_free_id();
		break;

		case '2':
		fprintf(stderr,"%s: Scanning for new/changed files.\n",progname);
		load_free_id();
		init_big_file();
		load_files();
		close_big_file();
		save_free_id();
		break;

		case '3':
		case '4':
		if (argc<3){
			gdbm_close(dbf);
			fprintf(stderr,"%s: Error: with switchcode 3 or 4 there"
				" must be another parametr as directory name\n",
				progname);
			exit(6);
		}
		{
			/* Let's look up this_directory_id */
			unsigned char *p;
			int len;
			datum key;
			unsigned long time;

			len=strlen(argv[2]);
			p=mem_alloc(len+1);
			strcpy(p,argv[2]);
			if (p[len-1]=='\n') p[(len--)-1]=0;
			key.dptr=p;
			key.dsize=len;
			parse_record(&this_directory_id, &time, key);
			free(p);
			
		}
		this_directory_index_filename=ssprintf("temptns/%lx.html",
			this_directory_id);	
		process_subentries(switchcode-'3');
		free(this_directory_index_filename);
		break;

		default:
		fprintf(stderr,"Function not implemented\n");
		break;
	}
	gdbm_close(dbf);
	return 0;
}
